@@ -7,7 +7,9 @@ from ..endpoint import Endpoint, EndpointClosedException
 
 
 UDPv4Address = collections.namedtuple("UDPv4Address", ["ip", "port"])
+UDPv4LANAddress = collections.namedtuple("UDPv4LANAddress", ["ip", "port"])
 UDPv6Address = collections.namedtuple("UDPv6Address", ["ip", "port"])
+DomainAddress = collections.namedtuple("DomainAddress", ["host", "port"])
 
 
 class UDPEndpoint(Endpoint, asyncio.DatagramProtocol):
@@ -133,4 +135,4 @@ class UDPv6Endpoint(UDPEndpoint):
         # If the endpoint is still running, accept incoming requests, otherwise drop them
         if self._running:
             self.bytes_down += len(datagram)
-            self.notify_listeners((UDPv6Address(*addr), datagram))
+            self.notify_listeners((UDPv6Address(*addr[:2]), datagram))
