@@ -1,7 +1,20 @@
+from __future__ import annotations
+
 import abc
+import socket
+import typing
 from binascii import hexlify
 from struct import Struct, pack, unpack_from
 
+from .interfaces.udp.endpoint import DomainAddress, UDPv4Address, UDPv6Address
+
+ADDRESS_TYPE_IPV4 = 0x01
+ADDRESS_TYPE_DOMAIN_NAME = 0x02
+ADDRESS_TYPE_IPV6 = 0x03
+
+
+FormatListType = typing.Union[str, "Serializable", typing.List["FormatListType"]]  # type:ignore
+
 
 class PackError(RuntimeError):
     pass
@@ -54,19 +67,19 @@ class NestedPayload(object):
         data = pack('>H', len(data)) + data
         return data
 
-    def unpack(self, serializable_class, data, offset, unpack_list):
+    def unpack(self, data, offset, unpack_list, serializable_class):
         """
         Unpack a Serializable using a class definition for some given data and offset.
         This is a special unpack_from which also takes a payload class.
 
-        :param serializable_class: the Serializable class to unpack to
-        :type serializable_class: type(Serializable)
         :param data: the data to unpack from
         :type data: str
         :param offset: the offset in the list of data to unpack from
         :type offset: int
         :param unpack_list: the list to which to append the Serializable
         :type unpack_list: list
+        :param serializable_class: the Serializable class to unpack to
+        :type serializable_class: type(Serializable)
         :return: the new offset
         :rtype: int
         """
@@ -85,14 +98,14 @@ class Bits(object):
         :type *data: list of 8 True or False values (or anything that maps to it in an if-statement)
         """
         byte = 0
-        byte |= 0x80 if data[7] else 0x00
-        byte |= 0x40 if data[6] else 0x00
-        byte |= 0x20 if data[5] else 0x00
-        byte |= 0x10 if data[4] else 0x00
-        byte |= 0x08 if data[3] else 0x00
-        byte |= 0x04 if data[2] else 0x00
-        byte |= 0x02 if data[1] else 0x00
-        byte |= 0x01 if data[0] else 0x00
+        byte |= 0x80 if data[0] else 0x00
+        byte |= 0x40 if data[1] else 0x00
+        byte |= 0x20 if data[2] else 0x00
+        byte |= 0x10 if data[3] else 0x00
+        byte |= 0x08 if data[4] else 0x00
+        byte |= 0x04 if data[5] else 0x00
+        byte |= 0x02 if data[6] else 0x00
+        byte |= 0x01 if data[7] else 0x00
         return pack('>B', byte)
 
     def unpack(self, data, offset, unpack_list):
@@ -129,7 +142,7 @@ class Raw(object):
 
 class VarLen(object):
     """
-    Paste/unpack from an encoded length + data string.
+    Pack/unpack from an encoded length + data string.
     """
 
     def __init__(self, length_format, base=1):
@@ -146,6 +159,75 @@ class VarLen(object):
         return offset + self.length_size + str_length
 
 
+class IPv4:
+    """
+    Pack/unpack an IPv4 address
+    """
+
+    def pack(self, data):
+        return pack('>4sH', socket.inet_aton(data[0]), data[1])
+
+    def unpack(self, data, offset, unpack_list):
+        host_bytes, port = unpack_from('>4sH', data, offset)
+        unpack_list.append(UDPv4Address(socket.inet_ntoa(host_bytes), port))
+        return offset + 6
+
+
+class Address:
+
+    def __init__(self, ip_only=False):
+        self.ip_only = ip_only
+
+    def pack(self, address):
+        if isinstance(address, UDPv6Address):
+            return pack('>B16sH', ADDRESS_TYPE_IPV6, socket.inet_pton(socket.AF_INET6, address.ip), address.port)
+        if not self.ip_only and isinstance(address, DomainAddress):
+            host_bytes = address.host.encode()
+            return pack(f'>BH{len(host_bytes)}sH', ADDRESS_TYPE_DOMAIN_NAME, len(host_bytes), host_bytes, address.port)
+        if isinstance(address, tuple):
+            return pack('>B4sH', ADDRESS_TYPE_IPV4, socket.inet_pton(socket.AF_INET, address[0]), address[1])
+        raise PackError(f'Unexpected address type {address}')
+
+    def unpack(self, data, offset, unpack_list):
+        address_type, = unpack_from('>B', data, offset)
+        if address_type == ADDRESS_TYPE_IPV4:
+            ip_bytes, port = unpack_from('>4sH', data, offset + 1)
+            unpack_list.append(UDPv4Address(socket.inet_ntop(socket.AF_INET, ip_bytes), port))
+            return offset + 7
+        elif address_type == ADDRESS_TYPE_IPV6:
+            ip_bytes, port = unpack_from('>16sH', data, offset + 1)
+            unpack_list.append(UDPv6Address(socket.inet_ntop(socket.AF_INET6, ip_bytes), port))
+            return offset + 19
+        elif not self.ip_only and address_type == ADDRESS_TYPE_DOMAIN_NAME:
+            length, = unpack_from('>H', data, offset + 1)
+            host = data[offset + 3: offset + 3 + length].decode()
+            unpack_list.append(DomainAddress(host, unpack_from('>H', data, offset + 3 + length)[0]))
+            return offset + 5 + length
+        else:
+            raise PackError(f'Cannot unpack address type {address_type}')
+
+
+class ListOf:
+
+    def __init__(self, packer, length_format='>B'):
+        self.packer = packer
+        self.length_format = length_format
+        self.length_size = Struct(length_format).size
+
+    def pack(self, data):
+        return pack(self.length_format, len(data)) + b''.join([self.packer.pack(item) for item in data])
+
+    def unpack(self, data, offset, unpack_list, *args):
+        length, = unpack_from(self.length_format, data, offset)
+        offset += self.length_size
+
+        result = []
+        unpack_list.append(result)
+        for _ in range(length):
+            offset = self.packer.unpack(data, offset, result, *args)
+        return offset
+
+
 class DefaultStruct:
 
     def __init__(self, format_str):
@@ -178,6 +260,7 @@ class Serializer(object):
             'I': DefaultStruct(">I"),
             'l': DefaultStruct(">l"),
             'LL': DefaultStruct(">LL"),
+            'q': DefaultStruct(">q"),
             'Q': DefaultStruct(">Q"),
             'QH': DefaultStruct(">QH"),
             'QL': DefaultStruct(">QL"),
@@ -190,13 +273,18 @@ class Serializer(object):
             '74s': DefaultStruct(">74s"),
             'c20s': DefaultStruct(">c20s"),
             'bits': Bits(),
+            'ipv4': IPv4(),
+            'ip_address': Address(ip_only=True),
+            'address': Address(),
             'raw': Raw(),
             'varlenBx2': VarLen('>B', 2),
             'varlenH': VarLen('>H'),
             'varlenHx20': VarLen('>H', 20),
+            'varlenH-list': ListOf(VarLen('>H')),
             'varlenI': VarLen('>I'),
             'doublevarlenH': VarLen('>H'),
-            'payload': NestedPayload(self)
+            'payload': NestedPayload(self),
+            'payload-list': ListOf(NestedPayload(self))
         }
 
     def get_available_formats(self):
@@ -211,14 +299,41 @@ class Serializer(object):
         """
         return self._packers[name]
 
-    def add_packing_format(self, name, format):
+    def add_packer(self, name, packer):
         """
-        Register a new struct packing format with a certain name.
+        Register a new packer with a certain name.
 
         :param name: the name to register
-        :param format: the format to use for it
+        :param packer: the packer to use for it
         """
-        self._packers.update({name: DefaultStruct(format)})
+        self._packers[name] = packer
+
+    def pack(self, fmt, item):
+        """
+        Pack data without using a Serializable. Using a Serializable is the preferred method.
+        :param fmt: the name of the packer to use while packing
+        :param item: object to pack
+        :return: the packed data
+        :rtype: bytes
+        """
+        return self._packers[fmt].pack(item)
+
+    def unpack(self, fmt, data, offset=0):
+        """
+        Unpack data without using a Serializable. Using a Serializable is the preferred method.
+        :param fmt: the name of the packer to use while unpacking
+        :param data: bytes to unpack
+        :return: the unpacked object
+        :rtype: object
+        """
+        unpack_list = []
+        if isinstance(fmt, str):
+            offset = self._packers[fmt].unpack(data, offset, unpack_list)
+        elif isinstance(fmt, list):
+            offset = self._packers['payload-list'].unpack(data, offset, unpack_list, fmt[0])
+        else:
+            offset = self._packers['payload'].unpack(data, offset, unpack_list, fmt)
+        return unpack_list[0], offset
 
     def pack_serializable(self, serializable):
         """
@@ -263,7 +378,11 @@ class Serializer(object):
             except KeyError:
                 if not issubclass(fmt, Serializable):
                     raise
-                offset = self._packers['payload'].unpack(fmt, data, offset, unpack_list)
+                offset = self._packers['payload'].unpack(data, offset, unpack_list, fmt)
+            except TypeError:
+                if not isinstance(fmt, list):
+                    raise
+                offset = self._packers['payload-list'].unpack(data, offset, unpack_list, fmt[0])
             except Exception as e:
                 raise PackError("Could not unpack item: %s\n%s: %s" % (fmt, type(e).__name__, str(e))) from e
         return serializable.from_unpack_list(*unpack_list), offset
@@ -301,10 +420,10 @@ class Serializable(metaclass=abc.ABCMeta):
     Interface for serializable objects.
     """
 
-    format_list = []
+    format_list: typing.List[FormatListType] = []
 
     @abc.abstractmethod
-    def to_pack_list(self):
+    def to_pack_list(self) -> typing.List[tuple]:
         """
         Serialize this object to a Serializer pack list.
 
@@ -314,13 +433,24 @@ class Serializable(metaclass=abc.ABCMeta):
         pass
 
     @abc.abstractmethod
-    def from_unpack_list(cls, *args):
+    def from_unpack_list(cls, *args) -> Serializable:  # pylint: disable=E0213
         """
         Create a new Serializable object from a list of unpacked variables.
         """
         pass
 
 
+class Payload(Serializable, abc.ABC):
+
+    def __str__(self):
+        out = self.__class__.__name__
+        for attribute in dir(self):
+            if not (attribute.startswith('_') or callable(getattr(self, attribute))) \
+                    and attribute not in ['format_list', 'names']:
+                out += '\n| %s: %s' % (attribute, repr(getattr(self, attribute)))
+        return out
+
+
 # Serializers should be stateless.
 # Therefore we can expose a global singleton for efficiency.
 # If you do need a Serializer with a state, be sure to use your own instance.
