@@ -4,7 +4,8 @@ import sys
 import threading
 import time
 import uuid
-from asyncio import all_tasks, get_event_loop, sleep
+from asyncio import all_tasks, ensure_future, get_event_loop, iscoroutine, sleep
+from functools import partial
 
 import asynctest
 
@@ -15,6 +16,20 @@ from ..peer import Peer
 get_event_loop().set_debug(True)
 
 
+def _on_packet_fragile_cb(self, packet, warn_unknown=True):
+    """
+    A fragile version of on_packet that crashes on message handling failures.
+
+    These failures won't actually cause IPv8 to crash in production, but you should probably handle these.
+
+    Add overlay classes to use in production mode to the ``production_overlay_classes`` list.
+    Filter nodes to run in production mode by overwriting ``TestBase.patch_overlays``.
+    """
+    result = self.decode_map[packet[1][22]](*packet)
+    if iscoroutine(result):
+        self.register_anonymous_task('on_packet', ensure_future(result))
+
+
 class TestBase(asynctest.TestCase):
 
     __testing__ = True
@@ -29,12 +44,14 @@ class TestBase(asynctest.TestCase):
         self.overlay_class = object
         internet.clear()
         self._tempdirs = []
+        self.production_overlay_classes = []
+        self._uncaught_async_failure = None
 
     def initialize(self, overlay_class, node_count, *args, **kwargs):
         self.overlay_class = overlay_class
         self.nodes = [self.create_node(*args, **kwargs) for _ in range(node_count)]
 
-        # Add nodes to each other
+        # Add nodes to each other.
         for node in self.nodes:
             for other in self.nodes:
                 if other == node:
@@ -44,18 +61,57 @@ class TestBase(asynctest.TestCase):
                 node.network.add_verified_peer(public_peer)
                 node.network.discover_services(public_peer, [overlay_class.community_id])
 
+        # Make packet handling fragile.
+        for i in range(len(self.nodes)):
+            self.patch_overlays(i)
+
+    def _patch_overlay(self, overlay):
+        if overlay and overlay.__class__ not in self.production_overlay_classes:
+            overlay.on_packet = partial(_on_packet_fragile_cb, overlay)
+
+    def patch_overlays(self, i):
+        """
+        Method to make the packet handlers of a particular node fragile.
+
+        If you want to disable fragile packet handling for an entire class over overlays, add the class to the
+        production_overlay_classes list.
+
+        If you want to disable fragile packet handling for a particular node, overwrite this method.
+        """
+        self._patch_overlay(self.node(i).overlay)
+        self._patch_overlay(self.node(i).dht)
+        for overlay in self.node(i).overlays:
+            self._patch_overlay(overlay)
+
+    def _cb_exception(self, loop, context):
+        """
+        Callback for asyncio exceptions.
+
+        Do not call `self.fail` or `pytest.fail` or any other failure mechanism in here.
+        These methods work by raising an Exception, which will silently fail as this is not the main loop.
+        """
+        # Only fail on the first failure, to not cause exception shadowing on compound errors.
+        if self._uncaught_async_failure is None:
+            self._uncaught_async_failure = context
+
     def setUp(self):
+        self.loop.set_exception_handler(self._cb_exception)
         super(TestBase, self).setUp()
         TestBase.__lockup_timestamp__ = time.time()
 
     async def tearDown(self):
         try:
             for node in self.nodes:
-                await node.unload()
+                await node.stop()
             internet.clear()
         finally:
             while self._tempdirs:
                 shutil.rmtree(self._tempdirs.pop(), ignore_errors=True)
+        # Now that everyone has calmed down, sweep up the remaining callbacks and check if they failed.
+        asynctest.helpers.exhaust_callbacks(self.loop)
+        if self._uncaught_async_failure is not None:
+            raise self._uncaught_async_failure["exception"]
+        self.loop.set_exception_handler(None)  # None is equivalent to the default handler
         super(TestBase, self).tearDown()
 
     @classmethod
@@ -136,7 +192,7 @@ class TestBase(asynctest.TestCase):
         rtime = 0
         probable_exit = False
 
-        while (rtime < timeout):
+        while rtime < timeout:
             await sleep(.01)
             rtime += .01
             if len([task for task in all_tasks() if not self.is_background_task(task)]) < 2:
@@ -159,3 +215,39 @@ class TestBase(asynctest.TestCase):
         self._tempdirs.append(d)
         os.makedirs(d)
         return d
+
+    def address(self, i):
+        return self.peer(i).address
+
+    def endpoint(self, i):
+        return self.nodes[i].endpoint
+
+    def key_bin(self, i):
+        return self.nodes[i].my_peer.public_key.key_to_bin()
+
+    def key_bin_private(self, i):
+        return self.nodes[i].my_peer.key.key_to_bin()
+
+    def mid(self, i):
+        return self.nodes[i].my_peer.mid
+
+    def my_peer(self, i):
+        return self.nodes[i].overlay.my_peer
+
+    def network(self, i):
+        return self.nodes[i].network
+
+    def node(self, i):
+        return self.nodes[i]
+
+    def overlay(self, i):
+        return self.nodes[i].overlay
+
+    def peer(self, i):
+        return Peer(self.nodes[i].my_peer.public_key.key_to_bin(), self.nodes[i].endpoint.wan_address)
+
+    def private_key(self, i):
+        return self.nodes[i].my_peer.key
+
+    def public_key(self, i):
+        return self.nodes[i].my_peer.public_key
