@@ -1,34 +1,25 @@
 import time
 
-from ..base import TestBase
-from ..mocking.ipv8 import MockIPv8
+from .base import TestDHTBase
 from ...dht.churn import PingChurn
 from ...dht.community import DHTCommunity
 
 
-class TestPingChurn(TestBase):
+class TestPingChurn(TestDHTBase):
 
     def setUp(self):
         super(TestPingChurn, self).setUp()
-
-        self.overlays = []
-        self.strategies = []
-
         self.initialize(DHTCommunity, 2)
 
-    def create_node(self, *args, **kwargs):
-        peer = MockIPv8(u"low", DHTCommunity)
-
-        self.overlays.append(peer.overlay)
-        self.strategies.append(PingChurn(peer.overlay, ping_interval=0.0))
-
-        return peer
+        self.strategies = []
+        for i in range(2):
+            self.strategies.append(PingChurn(self.overlay(i), ping_interval=0.0))
 
     async def test_ping_all(self):
         await self.introduce_nodes()
-        bucket = self.overlays[0].routing_table.trie[u'']
+        bucket = self.routing_table(0).trie[u'']
 
-        node1 = bucket.get(self.overlays[1].my_node_id)
+        node1 = bucket.get(self.my_node_id(1))
         node1.failed = 1
         node1.last_response = 0
 
@@ -40,8 +31,8 @@ class TestPingChurn(TestBase):
 
     async def test_ping_all_skip(self):
         await self.introduce_nodes()
-        bucket = self.overlays[0].routing_table.trie[u'']
-        node1 = bucket.get(self.overlays[1].my_node_id)
+        bucket = self.routing_table(0).trie[u'']
+        node1 = bucket.get(self.my_node_id(1))
         node1.failed = 1
         node1.last_response = time.time() + 5
 
