@@ -15,6 +15,10 @@ class TestEdgeWalk(TestBase):
         self.overlays = [MockCommunity() for _ in range(node_count)]
         self.strategies = [EdgeWalk(self.overlays[i], neighborhood_size=1) for i in range(node_count)]
 
+        # Prevent LAN IPs from ending up in Network
+        for overlay in self.overlays:
+            overlay.address_is_lan = lambda _: False
+
     async def tearDown(self):
         for overlay in self.overlays:
             await overlay.unload()
@@ -28,9 +32,9 @@ class TestEdgeWalk(TestBase):
           NODE0 <-> NODE1 <-> NODE2
         """
         self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
-        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].master_peer.mid, ])
+        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])
         self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
-        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].master_peer.mid, ])
+        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])
         # We expect NODE1 to introduce NODE0 to NODE2
         self.strategies[0].take_step()  # First it's added in the neighborhood
         self.strategies[0].take_step()  # Second it introduces its neighbor
@@ -48,7 +52,7 @@ class TestEdgeWalk(TestBase):
         """
         self.strategies[0].edge_timeout = 0.0  # Finish the edge immediately
         self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
-        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].master_peer.mid, ])
+        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])
 
         # We expect NODE0 will add NODE1 to its neighborhood and start constructing an edge from it.
         # Don't allow that right now.
@@ -58,7 +62,7 @@ class TestEdgeWalk(TestBase):
 
         # Now we give NODE2 to NODE1, which it can forward to NODE0 to make an edge
         self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
-        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].master_peer.mid, ])
+        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])
 
         # In order:
         # 1. Add root (NODE1) and query for nodes
@@ -89,7 +93,7 @@ class TestEdgeWalk(TestBase):
 
         # Now we give NODE2 to NODE1, which it can forward to NODE0 to make an edge
         self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
-        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].master_peer.mid, ])
+        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])
 
         # In order:
         # 1. Add root (NODE1) and query for nodes
@@ -109,7 +113,7 @@ class TestEdgeWalk(TestBase):
         self.strategies[0].edge_length = 2  # Finish with one other node
         self.strategies[0].edge_timeout = 1.0  # Finish the edge, but allow the network to walk
         self.overlays[0].network.add_verified_peer(self.overlays[1].my_peer)
-        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].master_peer.mid, ])
+        self.overlays[0].network.discover_services(self.overlays[1].my_peer, [self.overlays[1].community_id, ])
 
         # We expect NODE0 will add NODE1 to its neighborhood and start constructing an edge from it.
         # Don't allow that right now.
@@ -119,7 +123,7 @@ class TestEdgeWalk(TestBase):
 
         # Now we give NODE2 to NODE1, which it can forward to NODE0 to make an edge
         self.overlays[1].network.add_verified_peer(self.overlays[2].my_peer)
-        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].master_peer.mid, ])
+        self.overlays[1].network.discover_services(self.overlays[2].my_peer, [self.overlays[2].community_id, ])
 
         # In order:
         # 1. Add root (NODE1) and query for nodes
