@@ -16,6 +16,7 @@ def _generate_peer():
 class TestNetwork(TestBase):
 
     peers = [_generate_peer() for _ in range(4)]
+    serialized_ipv4_len = 7
 
     def setUp(self):
         super(TestNetwork, self).setUp()
@@ -137,7 +138,7 @@ class TestNetwork(TestBase):
         """
         Check if services are properly registered for a peer.
         """
-        service = "".join([chr(i) for i in range(20)])
+        service = bytes(range(20))
         self.network.discover_services(self.peers[0], [service])
         self.network.add_verified_peer(self.peers[0])
 
@@ -148,7 +149,7 @@ class TestNetwork(TestBase):
         """
         Check if services are properly registered for an unverified peer.
         """
-        service = "".join([chr(i) for i in range(20)])
+        service = bytes(range(20))
         self.network.discover_services(self.peers[0], [service])
 
         self.assertIn(service, self.network.get_services_for_peer(self.peers[0]))
@@ -158,8 +159,8 @@ class TestNetwork(TestBase):
         """
         Check if services are properly combined for a peer.
         """
-        service1 = "".join([chr(i) for i in range(20)])
-        service2 = "".join([chr(i) for i in range(20, 40)])
+        service1 = bytes(range(20))
+        service2 = bytes(range(20, 40))
         self.network.discover_services(self.peers[0], [service1])
         self.network.discover_services(self.peers[0], [service2])
         self.network.add_verified_peer(self.peers[0])
@@ -173,8 +174,8 @@ class TestNetwork(TestBase):
         """
         Check if services are properly combined when discovered services overlap.
         """
-        service1 = "".join([chr(i) for i in range(20)])
-        service2 = "".join([chr(i) for i in range(20, 40)])
+        service1 = bytes(range(20))
+        service2 = bytes(range(20, 40))
         self.network.discover_services(self.peers[0], [service1])
         self.network.discover_services(self.peers[0], [service1, service2])
         self.network.add_verified_peer(self.peers[0])
@@ -184,12 +185,39 @@ class TestNetwork(TestBase):
         self.assertIn(self.peers[0], self.network.get_peers_for_service(service1))
         self.assertIn(self.peers[0], self.network.get_peers_for_service(service2))
 
+    def test_discover_services_address_update(self):
+        """
+        Check if an address update gets processed correctly.
+        """
+        service1 = bytes(range(20))
+        service2 = bytes(range(20, 40))
+
+        pk = self.peers[0].public_key
+        peer = Peer(pk, ('1.1.1.1', 1))
+        self.network.add_verified_peer(peer)
+        self.network.discover_services(peer, [service1])
+
+        peer = Peer(pk, ('1.1.1.1', 1))
+        self.network.add_verified_peer(peer)
+        self.network.discover_services(peer, [service2])
+
+        peer = Peer(pk, ('1.1.1.1', 2))
+        self.network.add_verified_peer(peer)
+        self.network.discover_services(peer, [service1])
+
+        peer = Peer(pk, ('1.1.1.1', 2))
+        self.network.add_verified_peer(peer)
+        self.network.discover_services(peer, [service2])
+
+        self.assertEqual(self.network.get_peers_for_service(service1)[0].address, peer.address)
+        self.assertEqual(self.network.get_peers_for_service(service2)[0].address, peer.address)
+
     def test_discover_services_cache(self):
         """
         Check if services cache updates properly.
         """
-        service1 = "".join([chr(i) for i in range(20)])
-        service2 = "".join([chr(i) for i in range(20, 40)])
+        service1 = bytes(range(20))
+        service2 = bytes(range(20, 40))
         self.network.reverse_service_cache_size = 1
         self.network.discover_services(self.peers[0], [service1])
         self.network.add_verified_peer(self.peers[0])
@@ -374,7 +402,7 @@ class TestNetwork(TestBase):
         """
         Check if we can retrieve walkable addresses by parent service id.
         """
-        service = "".join([chr(i) for i in range(20)])
+        service = bytes(range(20))
         self.network.discover_address(self.peers[2], self.peers[3].address)
         self.network.discover_address(self.peers[0], self.peers[1].address)
         self.network.discover_services(self.peers[0], [service])
@@ -389,7 +417,7 @@ class TestNetwork(TestBase):
             self.network.add_verified_peer(peer)
         snapshot = self.network.snapshot()
 
-        self.assertEqual(len(snapshot), 6 * len(self.peers))
+        self.assertEqual(len(snapshot), TestNetwork.serialized_ipv4_len * len(self.peers))
 
     def test_snapshot_verified_and_unverified(self):
         """
@@ -399,7 +427,7 @@ class TestNetwork(TestBase):
         self.network.discover_address(self.peers[0], self.peers[1].address)
         snapshot = self.network.snapshot()
 
-        self.assertEqual(len(snapshot), 6)
+        self.assertEqual(len(snapshot), TestNetwork.serialized_ipv4_len)
 
     def test_snapshot_only_unverified(self):
         """
@@ -407,7 +435,7 @@ class TestNetwork(TestBase):
         """
         self.network.blacklist_mids.append(self.peers[0].mid)
         for peer in self.peers[1:]:
-            self.network.discover_address(self.peers[0], peer)
+            self.network.discover_address(self.peers[0], peer.address)
         snapshot = self.network.snapshot()
 
         self.assertEqual(len(snapshot), 0)
@@ -430,7 +458,10 @@ class TestNetwork(TestBase):
             ('194.116.42.174', 21882),
             ('143.93.254.175', 41862)
         }
-        snapshot = unhexlify("36e23871e14a8f5dfeafa386c2742aae557a46c8b71ded8f")
+        snapshot = unhexlify("01" "36e23871" "e14a"
+                             "01" "46c8b71d" "ed8f"
+                             "01" "c2742aae" "557a"
+                             "01" "8f5dfeaf" "a386")
 
         self.network.load_snapshot(snapshot)
         peers = set(self.network.get_walkable_addresses())
@@ -450,9 +481,20 @@ class TestNetwork(TestBase):
         """
         Check if no peers are loaded from a malformed snapshot.
         """
-        snapshot = unhexlify("36e23871e14a8f5dfeafa386c2742aae557a46c8b71ded8f")
+        snapshot = unhexlify("0136e23871e14a0146c8b71ded8f01c2742aae557a018f5dfeafa386")
+
+        self.network.load_snapshot(snapshot[1:])
+        peers = set(self.network.get_walkable_addresses())
+
+        self.assertSetEqual(peers, set())
+
+    def test_load_snapshot_truncated(self):
+        """
+        Check if as many peers as possible are loaded from a truncated snapshot.
+        """
+        snapshot = unhexlify("0136e23871e14a0146c8b71ded8f01c2742aae557a018f5dfeafa386")
 
         self.network.load_snapshot(snapshot[:-1])
         peers = set(self.network.get_walkable_addresses())
 
-        self.assertSetEqual(peers, set())
+        self.assertEqual(3, len(peers))
